
import java.util.*;
import java.net.*;
import java.io.*;

public class Client extends Thread
{
      Socket soc;

      BufferedReader in;
      PrintWriter out;

      PlayGui contact;

      String host;
      int port;

      boolean done = false;

/*      private boolean avail = true; */
     
      public Client(String host, int port, PlayGui contact)
      {
	 this.contact = contact;
	 this.host = host;
	 this.port = port;
      }

      public void run()
      {
	 while(!done)
	 {
	    Utility.sleep(1000);
	    if(contact.connected)
	    {
	       // notifications
	       String nots = checkNotifications();
	       if(nots != null && nots.length() > 0)
	       {
		  StringTokenizer st = new StringTokenizer(nots, "\n");
		  while(st.hasMoreTokens())
		  {
		     String tok = st.nextToken();
		     if(tok.equals("Update queue"))
			contact.updateQueue();
		     else if(tok.equals("Update tags"))
			contact.updateTags();
		  }
	       }

	       //status
	       String status = getStatus();

	       if(status == null)
	       {
		  contact.connectionDown();
		  contact.reportError("Couldn't read status from server, disconnecting.");
	       }
	       else if(status.startsWith("No File"))
	       {
		  contact.status = PlayGui.NO_FILE;
		  contact.togglePauseResumeButton.setEnabled(false);
		  contact.togglePauseResumeButton.setText("Pause");
	       }
	       else
	       {
		  if(status.startsWith("Stopped"))
		  {
		     contact.status = PlayGui.STOPPED;
		     contact.togglePauseResumeButton.setText("Play ");
		  }
		  else if(status.startsWith("Playing"))
		  {
		     contact.status = PlayGui.PLAYING;
		     contact.togglePauseResumeButton.setText("Pause");
		  }
		  else if(status.startsWith("Paused"))
		  {
		     contact.status = PlayGui.PAUSED;
		     contact.togglePauseResumeButton.setText("Play ");
		  }
		  contact.togglePauseResumeButton.setEnabled(true);
	       }
	    }
	 }
      }

/*    private synchronized boolean canTalk()
      {
	 if(!avail)
	    return false;
	 avail = false;
	 return true;
      }
*/

/*    private void doneTalking()
      {
         avail = true;
      }
*/

/*    private void waitForAvailability()
      {
	 while(!canTalk())
            Utility.sleep(100);
      }
*/

      // returns null on success, an error message otherwise
      public synchronized String connect()
      {
	 soc = new Socket();

	 try
	 {
	    soc.connect(new InetSocketAddress(host, port));
	    
	    in = new BufferedReader (new InputStreamReader( soc.getInputStream() ));
	    out = new PrintWriter (new OutputStreamWriter (soc.getOutputStream() ), true);
	    
/*	    waitForAvailability(); */
	    String status = in.readLine();
/*	    doneTalking(); */

	    if(!status.startsWith("1 "))
	       return "Server: " + status.substring(2,status.length());
	 }
	 catch (IOException ioe)
	 {
	    return("Connection to server " + host + " failed.  Bummer dude.");
	 }

	 return null;
      }

      public void disconnect()
      {
	 try
	 {
	    soc.close();
	    contact.updateQueue();
	    contact.updateSongList(false);
	    contact.updateTags();
	 }
	 catch(IOException ioe)
	 {
	    if(contact.debug > 1)
	       System.out.println("Error in disconnecting from " + host + ".  Far out.");
	    contact.reportError("Can't disconnect from host.");
	 }
      }

      String nowPlaying()
      {
	 String n = lister("GetNowPlaying");
	 if(n != null)
	    return n.trim();
	 return null;
      }

      String checkNotifications()
      {
	 return lister("CheckNotifications");
      }

      String listSongs()
      {
	 return lister("ListSongs");
      }

      String listTags()
      {
	 return lister("ListTags");
      }

      String listTagsForSong(String song)
      {
	 return lister("ListTagsForSong " + song);
      }

      String listQueue()
      {
	 return lister("GetQueue");
      }
      String listHistory()
      {
	 return lister("GetHistory");
      }
      
      private synchronized String lister(String cmd)
      {
/*	 waitForAvailability(); */
	 out.println(cmd);

	 try 
	 {
	    StringBuffer ret = new StringBuffer();

	    String cur = in.readLine();
	    	    
	    while (cur!= null && !(cur.startsWith("0 ") || cur.startsWith("1 ")))
	    {
	       ret.append(cur  + "\n"); 
	       cur = in.readLine();
	    }

	    // cur now is a status code
	    
	    if(cur == null)
	    {
	       contact.connectionDown();
	       return "";
	    }

	    if(cur.startsWith("0 "))
	    {
	       contact.reportError("Server: \"" +
				   cur.substring(2,cur.length()) +
				   "\".  Disconnecting.");
	       contact.connectionDown();
	       return null;
	    }

/*	    doneTalking(); */

	    if(cur != null &&
	       !"1 No notifications".equals(cur) 
	       && !"1 Status sent".equals(cur)
	       && !cur.endsWith("listed") 
	       && !cur.endsWith("sent"))
	       contact.setStatus("Server: \"" + cur.substring(2,cur.length()) + "\"");

	    if(contact.debug > 2)
	       System.out.println("Returning: " + ret.toString() + "\nReturned");
						
	    return ret.toString();
	 }
	 catch (IOException ioe)
	 {
	    contact.connectionDown();
	    contact.reportError("Couldn't read from host.  Sorry.");
/*	    doneTalking(); */
	    return null;
	 }
      }

 
      String multStr(String s, int times)
      {
	 StringBuffer t = new StringBuffer();
	 for(int i = 0 ; i < times ; i++)
	    t.append(s);
	 return t.toString();
      }


      private synchronized void simpleCommand(String cmd)
      {
/*	 waitForAvailability(); */

	 out.println(cmd);
	 try
	 {
	    String r = in.readLine();
	    if(r == null || !(r.startsWith(" 0") || r.startsWith("1 ")))
	    {
	       contact.connectionDown();
	       contact.reportError("Server not following protocall.  Recieved \"" + 
				   r + "\"");
	    }
	    else if(r.startsWith("0 "))
	       contact.reportError("Server: \"" +
				   r.substring(2,r.length()));
	    else // r.startsWith("1 ")
	       contact.setStatus("Server: \"" + r.substring(2,r.length()) + "\"");
	 }
	 catch (IOException ioe)
	 {
	    contact.connectionDown();
	    contact.reportError("Couldn't read from host.  Sorry.\n(command \"" + cmd + "\" failed.)");
	 }
/*	 doneTalking();*/
      }

      void removeFromHistory(int idx)
      {
	 simpleCommand("RemoveFromHistory " + idx);
      }

      void removeFromToplay(int idx)
      {
	 simpleCommand("RemoveFromToplay " + idx);
      }

      void removeIndices(int[] idxs)
      {
	 StringBuffer sb = new StringBuffer();
	 for(int i = 0 ; i < idxs.length ; i++)
	    sb.append( (i==0 ? "" : " ") + idxs[i]);
	 simpleCommand("RemoveSongsByIndex " + sb);
      }

      void refreshSongList()
      {
	 simpleCommand("RefreshSongList");
      }

      void clearToplay()
      {
	 simpleCommand("ClearQueue");
      }

      void clearHistory()
      {
	 simpleCommand("ClearHistory");
      }

      void pausee()
      {
	 simpleCommand("Pause");
      }

      void resumee()
      {
	 simpleCommand("Resume");
      }

      void startt()
      {
	 simpleCommand("Start");
      }

      void stopp()
      {
	 simpleCommand("Stop");
      }

      
      void next()
      {
	 simpleCommand("NextSong");
      }

      void nextN(int n)
      {
	 simpleCommand("NextSong " + n);
      }

      void prev()
      {
	 simpleCommand("PreviousSong");
      }

      void prevN(int n)
      {
	 simpleCommand("PreviousSong " + n);
      }

      String getStatus()
      {
	 String n = lister("Status");
	 if(n == null)
	    return null;
	 return n.trim();
      }

      String tagSearch(String expr)
      {
	 return lister("TagSearch " + expr);	 
      }

      void queueUpMult(Object[] songs)
      {
	 for(int i = 0 ; i < songs.length ; i++)
	    queueUp((String)songs[i]);
      }

      void tagMult(Object[] songs, String tag)
      {
	 for(int i = 0 ; i < songs.length ; i++)
	    tag((String)songs[i],tag);
      }

      void untagMult(Object[] songs, String tag)
      {
	 for(int i = 0 ; i < songs.length ; i++)
	    untag((String)songs[i],tag);
      }

      void newTag(String tag)
      {
	 simpleCommand("NewTag " + tag);
      }

      void tag(String song, String tag)
      {
	 simpleCommand("Tag " + tag + " " + song);
      }

      void untag(String song, String tag)
      {
	 simpleCommand("Untag " + tag + " " + song);
      }

      void removeTag(String tag)
      {
	 simpleCommand("DeleteTag " + tag);
      }

      void queueUp(String song)
      {
	 simpleCommand("QueueUpSong " + song);
      }
}
