import java.util.*;
import java.net.*;
import java.io.*;

class ConnectionHandler extends Thread
{
      Socket tsock;
      PlayMusicD deamon;
      boolean connectionValidated = false;
      boolean done = false;

      long lastCommand = System.currentTimeMillis();

      BufferedReader br;
      PrintWriter pw;

      LinkedList notifications = new LinkedList();

      String helpMessage = 
	 "This is a music player deamon.  Valid commands are:\n" +
	 "  Pause\n" +
	 "  Resume\n" + 
         "  Stop\n" +
         "  Start\n" +
	 "  ClearQueue\n" +
	 "  QueueUpSong <song>\n" + 
	 "  Status\n" + 
         "  RefreshSongList\n" +
	 "  ListSongs\n" + 
	 "  ListTags\n" +
	 "  ListTagsForSong <song>\n" +
	 "  Tag <tag> <song>\n" +
	 "  Untag <tag> <song>\n" +
	 "  TagSearch <expr>\n" +
	 "  GetQueue\n" + 
	 "  KillDeamon\n" +
	 "  CheckNotifications\n" + 
	 "  KeepAlive\n" + 
	 "  RemoveFromHistory <idx>\n" +
	 "  RemoveFromToplay <idx>\n" +
	 "  RemoveSongsByIndex <idx1> <idx2> <idx3> ...\n" +
	 "  Unmute\n" + 
	 "  PreviousSong\n" + 
	 "  NextSong\n" + 
	 "  PreviousSong <jmp>\n" + 
	 "  NextSong <jmp>\n\n";

      public ConnectionHandler(PlayMusicD deamon, Socket tsock, String allow, String deny)
      {
	 this.tsock = tsock;
	 this.deamon = deamon;

	 connectionValidated = allowConnection(tsock.getInetAddress().toString(), allow, deny);

	 try
	 {
	    br = new BufferedReader(new InputStreamReader(tsock.getInputStream()));
	    
	    //the boolean is to turn on autoflush
            pw = new PrintWriter(tsock.getOutputStream(),true);
	 }
	 catch(IOException ioe)
	 {
	    System.err.println("Couldn't make buffered readers and writers : " + tsock.getInetAddress());
	    done = true;
	 }
      }

      public void tellDead()
      {
	 done = true;
      }

      public boolean hasExpired()
      {
	 // A five minute timeout
	 return System.currentTimeMillis() - lastCommand > 5 * 60 * 1000;	    
      }

      public boolean allowConnection(String addr, String allow, String deny)
      {

	 if (addr == null)
	 {
	    if(deamon.debug > 0)
	       System.out.println("Recieved invalid null IP");
	    return false;
	 }

	 while(addr.startsWith("/"))
	    addr = addr.substring(1, addr.length());

	 // allow connections if they are explicity allowed or not explicitly denied.
	 return isMatched(addr, allow) || (! isMatched(addr, deny));
      }

      public boolean isMatched (String addr, String filter)
      {
	 if (filter == null )
	    return false;

	 if(filter.equals(addr))
	    return true;

	 if(filter.equals("ALL"))
	    return true;

	 if(filter.equals("NONE"))
	    return false;

	 StringTokenizer st = new StringTokenizer(filter);
	 String curTok;
	 while(st.hasMoreTokens())
	 {
	    curTok = st.nextToken();
	    if (addr.equals(curTok))
	       return true;

	    if (curTok.indexOf("*") != -1)
	    {
	       StringTokenizer A = new StringTokenizer(addr,".");
	       StringTokenizer B = new StringTokenizer(curTok,".");
	       boolean matchedSoFar = true;
		  
	       String tmpTokA;
	       String tmpTokB;
	       for (int i = 0 ; i < 4 ; i++)
	       {
		  tmpTokA = A.nextToken();
		  tmpTokB = B.nextToken();
		  
		  matchedSoFar = matchedSoFar && tmpTokB.equals("*") ? true : tmpTokA.equals(tmpTokB);
	       }
	       if (matchedSoFar)
		  return true;
	    }
	 }
	    
	 return false;
      }

      public void notify(String s)
      {
	 notifications.addFirst(s);
      }

      public void run()
      {
	 if(!connectionValidated)
	 {
	    if (deamon.debug > 1)
	       System.out.println("Rejected connection from " + tsock.getInetAddress() );
	    tell("0 Connection rejected");
	    deamon.server.removeConnection(this);
	    return;
	 }
	 else
	    tell("1 Connection accepted");

	 while (! done)
	 {
	    try
	    {
	       while(!br.ready())
		  Utility.sleep(30);
	    }
	    catch(IOException ioe)
	    {
	       System.err.println("Can't check if the stream from " + tsock.getInetAddress() + " is ready\n");
	       done = true;
	    }

	    String recd = null;
	    try 
	    {
	       recd = br.readLine();
	    }
	    catch(IOException ioe)
	    {
	       System.err.println("Can't read from " + tsock.getInetAddress() + "\n");
	       done = true;
	    }

	    if(deamon.debug > 1 && recd != null)
	       System.out.println("  Recd: " + recd);


	    if(recd != null)
	    {
	       lastCommand = System.currentTimeMillis();
	       
	       if (recd.equals("Pause"))
		  tell(deamon.pause());
	       else if (recd.equals("Resume"))
		  tell(deamon.resume());

	       else if (recd.equals("ClearQueue"))
		  tell(deamon.clearQueue());
	       else if (recd.equals("ClearHistory"))
		  tell(deamon.clearHistory());
	       else if (recd.startsWith("RemoveFromHistory "))
		  tell(deamon.removeFromHistory(recd.substring("RemoveFromHistory ".length() ,recd.length())));
	       else if (recd.startsWith("RemoveFromToplay "))
		  tell(deamon.removeFromToplay(recd.substring("RemoveFromToplay ".length() ,recd.length())));
	       else if (recd.startsWith("RemoveSongsByIndex "))
		  tell(deamon.removeSongsByIndex(recd.substring("RemoveSongsByIndex ".length() ,recd.length())));


	       else if (recd.startsWith("QueueUpSong"))
		  tell(deamon.queueUpSong(recd.substring("QueueUpSong ".length(),recd.length())));
//	       else if (recd.startsWith("QueueUpPlaylist"))
//		  tell(deamon.queueUpPlaylist(recd.substring("QueueUpPlaylist ".length(),recd.length())));

	       else if (recd.equals("Status"))
		  tell(deamon.getStatus());

	       else if (recd.equals("RefreshSongList"))
		  tell(deamon.refreshSongList());

	       else if (recd.equals("ListSongs"))
		  tell(deamon.listSongs());
	       else if (recd.equals("ListPlaylists"))
		  tell(deamon.listTags());


	       else if(recd.equals("ListTags"))
		  tell(deamon.listTags());
	       else if(recd.startsWith("ListTagsForSong "))
		  tell(deamon.listTagsForSong(recd.substring("ListTagsForSong ".length(),recd.length())) );
	       else if(recd.startsWith("Tag "))
		  tell(deamon.tag(recd.substring("Tag ".length(),recd.length())));
	       else if(recd.startsWith("Untag "))
		  tell(deamon.untag(recd.substring("Untag ".length(),recd.length())));
	       else if(recd.startsWith("NewTag "))
		  tell(deamon.newTag(recd.substring("NewTag ".length(),recd.length())));
	       else if(recd.startsWith("DeleteTag "))
		  tell(deamon.deleteTag(recd.substring("DeleteTag ".length(),recd.length())));
	       else if(recd.startsWith("TagSearch "))
		  tell(deamon.tagSearch(recd.substring("TagSearch ".length(),recd.length())) );
	    

	       else if (recd.equals("GetQueue"))
		  tell(deamon.getQueue());
	       else if (recd.equals("GetHistory"))
		  tell(deamon.getHistory());
	       else if (recd.equals("GetNowPlaying"))
		  tell(deamon.getNowPlaying());

	       else if (recd.equals("NextSong"))
		  tell(deamon.nextSong());
	       else if (recd.equals("PreviousSong"))
		  tell(deamon.previousSong());

	       else if(recd.equals("Stop"))
		  tell(deamon.stopp());
	       else if(recd.equals("Start"))
		  tell(deamon.startt());

	       else if (recd.startsWith("NextSong "))
		  tell(deamon.nextSongN(recd.substring("NextSong ".length(),recd.length())));
	       else if (recd.startsWith("PreviousSong "))
		  tell(deamon.previousSongN(recd.substring("PreviousSong ".length(),recd.length())));

	       else if (recd.equals("KillDeamon"))
	       {
		  tell("1 I am slain.  You inconsiderate little...");
		  Utility.die("Killed by " + tsock.getInetAddress());
	       }
	       else if (recd.equals("KeepAlive"))
	       {
		  tell("1 Keeping alive");
	       } 
	       else if (recd.equals("CheckNotifications"))
	       {
		  if(notifications.isEmpty())
		     tell("1 No notifications");
		  else
		  {
		     // remove duplicates and sort
		     Iterator it = (new TreeSet(notifications)).iterator(); 
		     
		     while(it.hasNext())
			tell((String)it.next());
		     notifications.clear();

		     tell("1 Notifications sent");
		  }
	       }
	       else if(recd.equals("Help") || recd.equals("ListCommands"))
	       {
		  tell(helpMessage);
		  tell("1 You've been helped now.");
	       }
	       else 
		  tell("0 Not understood");
	    }
	 }

	 try
	 {
	    tsock.close();
	 }
	 catch(IOException ioe)
	 {
	    if (deamon.debug > 0)
	       System.err.println("Can't close connection with " + tsock.getInetAddress());
	 }
      }

      public void tell(String s)
      {
	 pw.println(s);

	 if(deamon.debug > 1)
	    System.out.println("  Sent: \n" + s);
      }    
}

